/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller-Konrad (fabian dot peller-konrad at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

// STD/STL


// Header
#include "Reader.h"

// ArmarX
#include <ArmarXCore/core/rapidxml/wrapper/RapidXmlReader.h>
#include <ArmarXCore/core/system/cmake/CMakePackageFinder.h>

#include <RobotAPI/libraries/aron/core/typereader/xml/Data.h>
#include <RobotAPI/libraries/aron/core/type/variant/Factory.h>


namespace armarx::aron::typereader::xml
{
    namespace fs = std::filesystem;

    namespace
    {
        /// Resolve a relative Package path. This may be removed in a newer version of aron
        std::optional<fs::path> resolveRelativePackagePath(const fs::path& path)
        {
            const std::string package = *path.begin();
            armarx::CMakePackageFinder finder(package);
            if (finder.packageFound())
            {
                for (const std::string& includePath : finder.getIncludePathList())
                {
                    fs::path absPath = includePath / path;
                    if (fs::is_regular_file(absPath))
                    {
                        // path is valid
                        return absPath;
                    }
                }
                return std::nullopt;
            }
            else
            {
                return std::nullopt;
            }
        }
    }

    void Reader::parseFile(const std::string& _filename)
    {
        std::string filename = _filename;
        // Handle C++ style includes like "<path/to/file>".
        if (!filename.empty() && filename.front() == '<' && filename.back() == '>')
        {
            filename = filename.substr(1, filename.size() - 2);
        }
        parseFile(std::filesystem::path(filename));
    }

    void Reader::parseFile(const std::filesystem::path& _file)
    {
        fs::path file = _file;

        RapidXmlReaderPtr reader = RapidXmlReader::FromFile(file.string());
        parse(reader, _file);
    }

    // private method reading nodes
    void Reader::parse(const RapidXmlReaderPtr& reader, const std::filesystem::path& filePath)
    {
        RapidXmlReaderNode root = reader->getRoot();

        // Check Header
        util::EnforceTagName(root, constantes::TYPE_DEFINITION_TAG);

        util::EnforceChildSizeSmallerEqual(root, 4);

        auto code_includes = util::GetFirstNodeWithTag(root, constantes::CODE_INCLUDES_TAG);
        auto aron_includes = util::GetFirstNodeWithTag(root, constantes::INCLUDES_TAG);
        auto generate_types = util::GetFirstNodeWithTag(root, constantes::GENERATE_TYPES_TAG);

        // 1. Check includes
        if (code_includes.has_value())
        {
            for (const auto& include : (*code_includes).nodes())
            {
                this->codeIncludes.push_back(readCodeInclude(include, filePath));
            }
        }

        // 2. Check AronIncludes
        if (aron_includes.has_value())
        {
            for (const auto& aronInclude : (*aron_includes).nodes())
            {
                this->aronIncludes.push_back(readAronInclude(aronInclude, filePath));
            }
        }

        // 3. Check GenerateTypes
        if (generate_types.has_value())
        {
            for (const auto& generateType : (*generate_types).nodes())
            {
                if (util::HasTagName(generateType, constantes::OBJECT_TAG))
                {
                    const auto nav = readGenerateObject(generateType);
                    generateObjects.push_back(factory.allGeneratedPublicObjects.at(nav->getObjectName()));
                    continue;
                }

                if (util::HasTagName(generateType, constantes::INT_ENUM_TAG))
                {
                    const auto nav = readGenerateIntEnum(generateType);
                    generateIntEnums.push_back(factory.allGeneratedPublicIntEnums.at(nav->getEnumName()));
                    continue;
                }
                throw error::ValueNotValidException("XMLReader", "parse", "Could not find a valid tag inside generatetypes", generateType.name());
            }
        }
        else
        {
            throw error::AronException(__PRETTY_FUNCTION__, "No generate types found in aron xml '" + filePath.string() + "'.");
        }
    }

    std::string Reader::readCodeInclude(const RapidXmlReaderNode& node, const std::filesystem::path& filePath)
    {
        util::EnforceTagName(node, constantes::INCLUDE_TAG);
        util::EnforceAttribute(node, constantes::INCLUDE_ATTRIBUTE_NAME);
        const std::string include = util::GetAttribute(node, constantes::INCLUDE_ATTRIBUTE_NAME);
        return include;
    }

    std::string Reader::readAronInclude(const RapidXmlReaderNode& node, const std::filesystem::path& filePath)
    {
        util::EnforceTagName(node, constantes::INCLUDE_TAG);
        std::string specifiedPath = util::GetAttribute(node, constantes::INCLUDE_ATTRIBUTE_NAME);
        specifiedPath = simox::alg::replace_all(specifiedPath, "<", "");
        specifiedPath = simox::alg::replace_all(specifiedPath, ">", "");
        const std::filesystem::path xmlincludepath(specifiedPath);

        // add logic to allow relative paths:
        std::filesystem::path resolved_absolute_path(xmlincludepath);
        std::filesystem::path resolved_relative_path(xmlincludepath);

        if (!xmlincludepath.empty() && xmlincludepath.is_relative())
        {
            if (std::optional<fs::path> resolvedPackagePath = resolveRelativePackagePath(xmlincludepath); resolvedPackagePath.has_value())
            {
                // check if relative path belongs to a package
                resolved_absolute_path = resolvedPackagePath.value();
                resolved_relative_path = xmlincludepath;
            }
            else
            {
                // perhaps the relative path is relative from file position?
                // TODO: FixMe
                resolved_absolute_path = filePath.parent_path() / xmlincludepath;
                resolved_relative_path = filePath.parent_path() / xmlincludepath;

                if (!fs::is_regular_file(resolved_absolute_path))
                {
                    throw error::AronException(__PRETTY_FUNCTION__, "Could not find an aron XML file. Last path tried was: " + resolved_absolute_path.string());
                }
            }
        }
        // else path is absolute

        // parse parent xml file and add objects to alreday known
        Reader anotherReader;
        anotherReader.parseFile(resolved_absolute_path);
        for (const auto& previouslyKnown : anotherReader.factory.allPreviouslyKnownPublicTypes)
        {
            factory.allPreviouslyKnownPublicTypes.push_back(previouslyKnown);
        }
        for (const auto& knownObjectInfo : anotherReader.factory.allGeneratedPublicObjects)
        {
            factory.allPreviouslyKnownPublicTypes.push_back(knownObjectInfo.first);
        }
        for (const auto& knownIntEnumInfo : anotherReader.factory.allGeneratedPublicIntEnums)
        {
            factory.allPreviouslyKnownPublicTypes.push_back(knownIntEnumInfo.first);
        }

        if (util::HasAttribute(node, constantes::AUTO_CODE_INCLUDE))
        {
            std::string codeinclude = simox::alg::replace_last(resolved_relative_path, ".xml", ".aron.generated.h");
            this->codeIncludes.push_back("<" + codeinclude + ">");
        }

        return resolved_absolute_path.string();
    }

    type::ObjectPtr Reader::readGenerateObject(const RapidXmlReaderNode& node)
    {
        util::EnforceTagName(node, constantes::OBJECT_TAG);
        return type::Object::DynamicCastAndCheck(factory.create(node, Path()));
    }

    type::IntEnumPtr Reader::readGenerateIntEnum(const RapidXmlReaderNode& node)
    {
        util::EnforceTagName(node, constantes::INT_ENUM_TAG);
        return type::IntEnum::DynamicCastAndCheck(factory.create(node, Path()));
    }
}


